import requests
import requests.packages
from typing import Dict
from .exceptions import OHGOException
from ohgo.models.models import Result
from json import JSONDecodeError
import logging
from io import BytesIO


class RestAdapter:
    """
    RestAdapter is a class for making HTTP requests to the OHGO API

    Attributes:
    url: The base URL of the OHGO API
    _api_key: The API key for the OHGO API
    _ssl_verify: Whether to verify SSL certificates
    _logger: A logger for logging messages

    Methods:
    get: Makes a GET request to the OHGO API
    get_image: Fetches an image from a URL
    _do: Makes a request to the OHGO API
    """
    def __init__(
            self,
            hostname: str,
            api_key: str = "",
            ver: str = "v1",
            ssl_verify: bool = True,
            logger: logging.Logger = None,
    ):
        """
        Constructor for RestAdapter. Initializes the base URL, API key, SSL verification, and logger.
        :param hostname: hostname of the OHGO API. Almost always "publicapi.ohgo.com"
        :param api_key: API key for the OHGO API
        :param ver: Version of the API to use. Defaults to "v1"
        :param ssl_verify: Whether to verify SSL certificates. Defaults to True
        :param logger: (optional) A logger to use for logging. Defaults to None. A new logger will be created if None.
        """

        self.url = "https://{}/api/{}/".format(hostname, ver)
        self._api_key = api_key
        self._ssl_verify = ssl_verify
        self._logger = logger or logging.getLogger(__name__)
        if not ssl_verify:
            requests.packages.urllib3.disable_warnings()

    def get(self, endpoint: str, ep_params: Dict = {}, fetch_all=False) -> Result:
        """
        Makes a GET request to the OHGO API
        :param endpoint: The endpoint to make the request to
        :param ep_params: The parameters to pass to the endpoint
        :param fetch_all: Whether to fetch all results. Defaults to False. Recommended to use page-all param instead.
        :return: A Result object
        """
        result = self._do(http_method="GET", endpoint=endpoint, ep_params=ep_params)
        if fetch_all:
            # Fetch all results by following the next page links
            next_page_url = result.next_page
            while next_page_url:
                page_result = self._do(http_method="GET", endpoint=next_page_url, ep_params=ep_params)
                result.data.extend(page_result.data)
                next_page_url = page_result.next_page
        return result

    def get_image(self, url) -> BytesIO:
        """
        Fetches an image from a URL
        :param url: The URL to fetch the image from
        :return: A BytesIO object containing the image
        """
        try:
            response = requests.get(url, verify=self._ssl_verify)
            response.raise_for_status()
            return BytesIO(response.content)
        except requests.RequestException as e:
            self._logger.error(f"Error while fetching image from {url}: {e}")
            raise OHGOException(f"Failed to fetch image from {url}") from e

    def _do(
            self, http_method: str, endpoint: str, ep_params: Dict = {}, data: Dict = {}
    ) -> Result:
        """
        Helper method that makes a request to the OHGO API
        :param http_method: The HTTP method to use. Currently, OHGO only supports GET
        :param endpoint: The endpoint to make the request to
        :param ep_params: The parameters to pass to the endpoint
        :param data: The data to pass to the endpoint.
        :return: A Result object
        """
        full_url = endpoint if endpoint.startswith('http') else self.url + endpoint
        headers = {
            "Authorization": f"APIKEY {self._api_key}"
        }
        ep_params = {k: v for k, v in ep_params.items() if v is not None}
        print(http_method, endpoint, ep_params, data)
        print(full_url)
        try:
            response = requests.request(
                method=http_method,
                url=full_url,
                verify=self._ssl_verify,
                headers=headers,
                params=ep_params,
                json=data,
            )
        except (ValueError, JSONDecodeError) as e:
            raise OHGOException("Request failed.") from e
        data_out = response.json()
        if 299 >= response.status_code >= 200:
            # Successful request
            result = Result(
                status_code=response.status_code,
                message=response.reason,
                data=data_out,
            )

            for query_filter in result.rejected_filters:
                # OHGO rejected a filter, log a warning
                self._logger.warning(f" Error: {query_filter['error']} - {query_filter['key']}:{query_filter['value']}")

            return result
        raise OHGOException(f"{response.status_code}: {response.reason}")



from dataclasses import dataclass
from datetime import datetime
from typing import List, Any

from ohgo.models.base_model import BaseModel
from ohgo.models.models import Link, from_list, from_str, from_float, from_datetime, to_float, to_class


@dataclass
class Construction(BaseModel):
    """
    Construction is a class for storing construction information.

    Attributes:
    category: The category of the construction work.
    direction: The direction related to the construction site.
    district: The district where the construction is taking place (can be None).
    route_name: The name of the route where the construction is located.
    status: The current status of the construction work.
    start_date: The start date of the construction work.
    end_date: The expected end date of the construction work.
    """

    category: str
    direction: str
    district: str
    route_name: str
    status: str
    start_date: datetime
    end_date: datetime

    def __init__(
            self,
            links: List[Link],
            id: str,
            latitude: float,
            longitude: float,
            location: str,
            description: str,
            category: str,
            direction: str,
            district: str,
            route_name: str,
            status: str,
            start_date: datetime,
            end_date: datetime
    ):
        """
        Initializes the Construction object.
        :param links: A list of Link objects associated with the construction work.
        :param id: The ID of the construction work.
        :param latitude: The latitude of the construction site.
        :param longitude: The longitude of the construction site.
        :param location: The location where the construction is happening.
        :param description: A description of the construction work.
        :param category: The category of the construction work.
        :param direction: The direction associated with the construction work.
        :param district: The district where the construction is located.
        :param route_name: The name of the route affected by the construction work.
        :param status: The current status of the construction work.
        :param start_date: The start date of the construction work.
        :param end_date: The expected end date of the construction work.
        """
        super().__init__(links, id, latitude, longitude, location, description)  # Initialize common fields from BaseModel
        self.category = category
        self.direction = direction
        self.district = district
        self.route_name = route_name
        self.status = status
        self.start_date = start_date
        self.end_date = end_date

    @staticmethod
    def from_dict(obj: Any) -> "Construction":
        """
        Converts a dictionary into a Construction object.
        :param obj: A dictionary representing a Construction object.
        :return: A Construction object.
        """
        base_model = BaseModel.from_base_dict(obj)  # Parse common fields using BaseModel
        category = from_str(obj.get("category"))
        direction = from_str(obj.get("direction"))
        district = from_str(obj.get("district"))
        route_name = from_str(obj.get("routeName"))
        status = from_str(obj.get("status"))
        start_date = from_datetime(obj.get("startDate"))
        end_date = from_datetime(obj.get("endDate"))
        return Construction(
            base_model.links, base_model.id, base_model.latitude, base_model.longitude, base_model.location,
            base_model.description, category, direction, district, route_name, status, start_date, end_date
        )

    def to_dict(self) -> dict:
        """
        Converts the Construction object into a dictionary.
        :return: A dictionary representation of the Construction object.
        """
        result = self.base_to_dict()  # Get common fields from BaseModel
        result.update({
            "category": from_str(self.category),
            "direction": from_str(self.direction),
            "district": from_str(self.district),
            "routeName": from_str(self.route_name),
            "status": from_str(self.status),
            "startDate": self.start_date.isoformat(),
            "endDate": self.end_date.isoformat()
        })
        return result




from dataclasses import dataclass
from typing import List, Any

from ohgo.models.base_model import BaseModel
from ohgo.models.models import from_list, from_str, from_float, to_class, to_float, Link


@dataclass
class CameraView:
    """
    CameraView is a class for storing the view of a camera.

    Attributes:
    direction: The direction the camera is facing
    small_url: The URL of the small image
    large_url: The URL of the large image
    main_route: The main route of the camera

    Methods:
    from_dict: Converts a dictionary to a CameraView object
    to_dict: Converts the CameraView object to a dictionary
    """
    direction: str
    small_url: str
    large_url: str
    main_route: str

    def __init__(self, direction, small_url, large_url, main_route):
        """
        Initializes the CameraView object with the direction, small URL, large URL, and main route.
        :param direction: The direction the camera is facing as a string, PTZ if the camera is not in a fixed position
        :param small_url: URL link to the small image. Image snapshots are updated every 5 seconds.
        :param large_url: URL link to the large image. Image snapshots are updated every 5 seconds.
        :param main_route: The main road/intersection the camera is monitoring
        """
        self.direction = direction
        self.small_url = small_url
        self.large_url = large_url
        self.main_route = main_route
        self._small_image_cache = None
        self._large_image_cache = None

    @staticmethod
    def from_dict(obj: Any) -> "CameraView":
        assert isinstance(obj, dict)
        direction = from_str(obj.get("direction"))
        small_url = from_str(obj.get("smallUrl"))
        large_url = from_str(obj.get("largeUrl"))
        main_route = from_str(obj.get("mainRoute"))
        return CameraView(direction, small_url, large_url, main_route)

    def to_dict(self) -> dict:
        result: dict = {"direction": from_str(self.direction), "smallUrl": from_str(self.small_url),
                        "largeUrl": from_str(self.large_url), "mainRoute": from_str(self.main_route)}
        return result

@dataclass
class Camera(BaseModel):
    """
    Camera is a class for storing camera objects.

    Attributes:
    camera_views: The views of the camera. Each view is a CameraView object that contains information about an image.

    """

    camera_views: List[CameraView]
    def __init__(
            self,
            links: List[Link],
            id: str,
            latitude: float,
            longitude: float,
            location: str,
            description: str,
            camera_views: List[CameraView]
    ):
        """
        Initializes the Camera object with common fields and camera views.
        :param links: A list of Link objects that relate to the Camera. Usually just a direct link to the camera's page.
        :param id: The ID of the camera.
        :param latitude: The latitude of the camera's location.
        :param longitude: The longitude of the camera's location.
        :param location: The location of the camera. Usually a road or intersection.
        :param description: The description of the camera's location.
        :param camera_views: A list of CameraView objects that contain information about the camera's views.
        """
        super().__init__(links, id, latitude, longitude, location, description)  # Call BaseModel's init
        self.camera_views = camera_views

    @staticmethod
    def from_dict(obj: Any) -> "Camera":
        """
        Converts a dictionary into a Camera object.
        :param obj: A dictionary representing a Camera object.
        :return: A Camera object.
        """
        base_model = BaseModel.from_base_dict(obj)  # Get common fields using BaseModel
        camera_views = from_list(CameraView.from_dict, obj.get("cameraViews"))
        return Camera(base_model.links, base_model.id, base_model.latitude, base_model.longitude, base_model.location, base_model.description, camera_views)

    def to_dict(self) -> dict:
        """
        Converts the Camera object into a dictionary.
        :return: A dictionary representation of the Camera object.
        """
        result = self.base_to_dict()  # Get common fields from BaseModel
        result.update({
            "cameraViews": from_list(lambda x: to_class(CameraView, x), self.camera_views)
        })
        return result



from dataclasses import dataclass
from datetime import datetime
from typing import List, Any, Optional

from ohgo.models.base_model import BaseModel
from ohgo.models.models import from_list, from_str, Link, from_float, to_class, to_float, from_bool, \
    from_datetime, from_int


@dataclass
class AtmosphericSensor:
    air_temperature: float
    dewpoint_temperature: float
    humidity: float
    average_wind_speed: float
    maximum_wind_speed: float
    wind_direction: str
    precipitation: str
    precipitation_rate: float
    visibility: float
    last_update: datetime

    @staticmethod
    def from_dict(obj: Any) -> 'AtmosphericSensor':
        assert isinstance(obj, dict)
        air_temperature = from_float(obj.get("airTemperature"))
        dewpoint_temperature = from_float(obj.get("dewpointTemperature"))
        humidity = from_float(obj.get("humidity"))
        average_wind_speed = from_float(obj.get("averageWindSpeed"))
        maximum_wind_speed = from_float(obj.get("maximumWindSpeed"))
        wind_direction = from_str(obj.get("windDirection"))
        precipitation = from_str(obj.get("precipitation"))
        precipitation_rate = from_float(obj.get("precipitationRate"))
        visibility = from_float(obj.get("visibility"))
        last_update = from_datetime(obj.get("lastUpdate"))
        return AtmosphericSensor(air_temperature, dewpoint_temperature, humidity, average_wind_speed,
                                 maximum_wind_speed, wind_direction, precipitation, precipitation_rate, visibility,
                                 last_update)

    def to_dict(self) -> dict:
        result: dict = {"airTemperature": to_float(self.air_temperature),
                        "dewpointTemperature": to_float(self.dewpoint_temperature), "humidity": to_float(self.humidity),
                        "averageWindSpeed": to_float(self.average_wind_speed),
                        "maximumWindSpeed": to_float(self.maximum_wind_speed),
                        "windDirection": from_str(self.wind_direction), "precipitation": from_str(self.precipitation),
                        "precipitationRate": to_float(self.precipitation_rate), "visibility": to_float(self.visibility),
                        "lastUpdate": self.last_update.isoformat()}
        return result


@dataclass
class SurfaceSensor:
    name: str
    status: str
    surface_temperature: float
    sub_surface_temperature: float
    last_update: datetime

    @staticmethod
    def from_dict(obj: Any) -> 'SurfaceSensor':
        assert isinstance(obj, dict)
        name = from_str(obj.get("name"))
        status = from_str(obj.get("status"))
        surface_temperature = from_float(obj.get("surfaceTemperature"))
        sub_surface_temperature = from_float(obj.get("subSurfaceTemperature"))
        last_update = from_datetime(obj.get("lastUpdate"))
        return SurfaceSensor(name, status, surface_temperature, sub_surface_temperature, last_update)

    def to_dict(self) -> dict:
        result: dict = {"name": from_str(self.name), "status": from_str(self.status),
                        "surfaceTemperature": to_float(self.surface_temperature),
                        "subSurfaceTemperature": to_float(self.sub_surface_temperature),
                        "lastUpdate": self.last_update.isoformat()}
        return result


@dataclass
class WeatherSensorSite(BaseModel):
    """
    WeatherSensorSite is a class for storing weather sensor site objects.

    Attributes:
    severe: A boolean indicating if the weather conditions are severe.
    condition: The current weather condition at the site.
    average_air_temperature: The average air temperature at the site.
    atmospheric_sensors: A list of atmospheric sensors at the site.
    surface_sensors: A list of surface sensors at the site.
    """

    severe: bool
    condition: Optional[str]
    average_air_temperature: str
    atmospheric_sensors: List[AtmosphericSensor]
    surface_sensors: List[SurfaceSensor]

    def __init__(
            self,
            links: List[Link],
            id: str,
            latitude: float,
            longitude: float,
            location: str,
            description: Optional[str],
            severe: bool,
            condition: Optional[str],
            average_air_temperature: str,
            atmospheric_sensors: List[AtmosphericSensor],
            surface_sensors: List[SurfaceSensor]
    ):
        """
        Initializes the WeatherSensorSite object with common fields and weather-specific fields.
        :param links: A list of Link objects associated with the weather sensor site.
        :param id: The ID of the weather sensor site.
        :param latitude: The latitude of the weather sensor site.
        :param longitude: The longitude of the weather sensor site.
        :param location: The location of the weather sensor site.
        :param description: A description of the weather sensor site.
        :param severe: A boolean indicating if the weather conditions are severe.
        :param condition: The current weather condition at the site.
        :param average_air_temperature: The average air temperature recorded at the site.
        :param atmospheric_sensors: A list of AtmosphericSensor objects.
        :param surface_sensors: A list of SurfaceSensor objects.
        """
        super().__init__(links, id, latitude, longitude, location, description)  # Call BaseModel's init
        self.severe = severe
        self.condition = condition
        self.average_air_temperature = average_air_temperature
        self.atmospheric_sensors = atmospheric_sensors
        self.surface_sensors = surface_sensors

    @staticmethod
    def from_dict(obj: Any) -> 'WeatherSensorSite':
        """
        Converts a dictionary into a WeatherSensorSite object.
        :param obj: A dictionary representing a WeatherSensorSite.
        :return: A WeatherSensorSite object.
        """
        base_model = BaseModel.from_base_dict(obj)  # Get common fields using BaseModel
        severe = from_bool(obj.get("severe"))
        condition = from_str(obj.get("condition"))
        average_air_temperature = from_str(obj.get("averageAirTemperature"))
        atmospheric_sensors = from_list(AtmosphericSensor.from_dict, obj.get("atmosphericSensors"))
        surface_sensors = from_list(SurfaceSensor.from_dict, obj.get("surfaceSensors"))
        return WeatherSensorSite(base_model.links, base_model.id, base_model.latitude, base_model.longitude,
                                 base_model.location, base_model.description, severe, condition, average_air_temperature,
                                 atmospheric_sensors, surface_sensors)

    def to_dict(self) -> dict:
        """
        Converts the WeatherSensorSite object into a dictionary.
        :return: A dictionary representation of the WeatherSensorSite object.
        """
        result = self.base_to_dict()  # Get common fields from BaseModel
        result.update({
            "severe": from_bool(self.severe),
            "condition": from_str(self.condition),
            "averageAirTemperature": from_str(self.average_air_temperature),
            "atmosphericSensors": from_list(lambda x: to_class(AtmosphericSensor, x), self.atmospheric_sensors),
            "surfaceSensors": from_list(lambda x: to_class(SurfaceSensor, x), self.surface_sensors)
        })
        return result




from dataclasses import dataclass
from typing import List, Any

from ohgo.models.base_model import BaseModel
from ohgo.models.models import from_list, from_str, from_float, to_float, to_class, Link


@dataclass
class DigitalSign(BaseModel):
    """
    DigitalSign is a class for storing digital sign objects.

    Attributes:
    sign_type_name: The name/type of the digital sign.
    messages: A list of messages displayed by the sign.
    image_urls: A list of URLs for images associated with the digital sign.
    """

    sign_type_name: str
    messages: List[str]
    image_urls: List[Any]

    def __init__(
            self,
            links: List[Link],
            id: str,
            latitude: float,
            longitude: float,
            location: str,
            description: str,
            sign_type_name: str,
            messages: List[str],
            image_urls: List[Any]
    ):
        """
        Initializes the DigitalSign object with common fields and sign-specific fields.
        :param links: A list of Link objects associated with the sign.
        :param id: The ID of the sign.
        :param latitude: The latitude of the sign's location.
        :param longitude: The longitude of the sign's location.
        :param location: The location of the sign.
        :param description: The description of the sign.
        :param sign_type_name: The name or type of the digital sign.
        :param messages: A list of messages displayed by the digital sign.
        :param image_urls: A list of image URLs associated with the digital sign.
        """
        super().__init__(links, id, latitude, longitude, location, description)  # Call BaseModel's init
        self.sign_type_name = sign_type_name
        self.messages = messages
        self.image_urls = image_urls

    @staticmethod
    def from_dict(obj: Any) -> 'DigitalSign':
        """
        Converts a dictionary into a DigitalSign object.
        :param obj: A dictionary representing a DigitalSign.
        :return: A DigitalSign object.
        """
        base_model = BaseModel.from_base_dict(obj)  # Get common fields using BaseModel
        sign_type_name = from_str(obj.get("signTypeName"))
        messages = from_list(from_str, obj.get("messages"))
        image_urls = from_list(lambda x: x, obj.get("imageUrls"))
        return DigitalSign(base_model.links, base_model.id, base_model.latitude, base_model.longitude,
                           base_model.location, base_model.description, sign_type_name, messages, image_urls)

    def to_dict(self) -> dict:
        """
        Converts the DigitalSign object into a dictionary.
        :return: A dictionary representation of the DigitalSign object.
        """
        result = self.base_to_dict()  # Get common fields from BaseModel
        result.update({
            "signTypeName": from_str(self.sign_type_name),
            "messages": from_list(from_str, self.messages),
            "imageUrls": from_list(lambda x: x, self.image_urls)
        })
        return result


